'''Selector functions and selector function factories.'''

__author__ = 'Robert Smallshire'


def k_(key):
    '''Create a selector function which indexes into the element by key.

    Args:
        key: The key which the generated selector will use to index into
            elements.

    Returns:
        A unary selector function which indexes into its only argument with
        the key value.
    '''
    return lambda element: element[key]


def a_(name):
    '''Create a selector function which selects an attribute by name.

    Args:
        name: The name of the attribute which will be retrieved from each
            element.

    Returns:
        A unary selector function which retrieves the named attribute from its
        only argument and returns the value of that attribute.
    '''
    return lambda element: getattr(element, name)


def m_(name, *args, **kwargs):
    '''Create a selector function which calls a named method.

    Args:
        name: The name of the method which will be called on each element.

        *args: Any optional positional arguments which will be passed to the
            called method.

        **kwargs: Any optional named arguments which will be passed to the
            called method.

    Returns:
        A unary selector function which calls the named method with any
        optional positional or named arguments and which returns the
        result of the method call.
    '''
    return lambda element: getattr(element, name)(*args, **kwargs)


def identity(x):
    '''The identity function.

    The identity function returns its only argument.

    Args:
        x: A value that will be returned.

    Returns:
        The argument x.
    '''
    return x



__author__ = 'Robert Smallshire'


def insensitive_eq(lhs, rhs):
    '''Case insensitive string equality operator.

    Args:
        lhs (str): The first string to be compared.
        rhs (str): The second string to be compared.

    Returns:
        True if the strings are equal in a case insensitive way, otherwise
        False.
    '''
    return lhs.lower() == rhs.lower()



'''Initiators are factory functions for creating Queryables.'''
import itertools

from ._portability import irange

__author__ = 'Robert Smallshire'


def query(iterable):
    '''Make an iterable queryable.

    Use this function as an entry-point to the asq system of chainable query
    methods.

    Note: Currently this factory only provides support for objects supporting
        the iterator protocol.  Future implementations may support other
        providers.

    Args:
        iterable: Any object supporting the iterator protocol.

    Returns:
        An instance of Queryable.

    Raises:
        TypeError: If iterable is not actually iterable
    '''
    # Avoid a circular module dependency
    from .queryables import Queryable
    return Queryable(iterable)


def integers(start, count):
    '''Generates in sequence the integral numbers within a range.

    Note: This method uses deferred execution.

    Args:
        start: The first integer in the sequence.
        count: The number of sequential integers to generate.

    Returns:
        A Queryable over the specified range of integers.

    Raises:
        ValueError: If count is negative.
    '''
    if count < 0:
        raise ValueError("integers() count cannot be negative")
    return query(irange(start, start + count))


def repeat(element, count):
    '''Generate a sequence with one repeated value.

    Note: This method uses deferred execution.

    Args:
        element: The value to be repeated.
        count: The number of times to repeat the value.

    Raises:
        ValueError: If the count is negative.
    '''
    if count < 0:
        raise ValueError("repeat() count cannot be negative")
    return query(itertools.repeat(element, count))


_empty = None


def empty():
    '''An empty Queryable.

    Note: The same empty instance will be returned each time.

    Returns: A Queryable over an empty sequence.
    '''
    # We use this lazy initialization of _empty to avoid a circular
    # module dependency
    global _empty
    if _empty is None:
        _empty = query(tuple())
    return _empty



'''Python 2 and Python 3 compatibility'''

__author__ = 'rjs'

import sys

try:
    # Python 2
    from itertools import imap
except ImportError:
    # Python 3
    imap = map

try:
    # Python 2
    from itertools import ifilter
except ImportError:
    # Python 3
    ifilter = filter

try:
    # Python 2
    from itertools import izip
except ImportError:
    # Python 3
    izip = zip

try:
    # Python 2
    from itertools import izip_longest
except ImportError:
    # Python 3
    from itertools import zip_longest as izip_longest

try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError:
        sys.stderr.write('Could not import OrderedDict. For Python versions '
                         'earlier than 2.7 install the')
        sys.stderr.write('ordereddict module from the Python Package Index '
                         'with easy_install ordereddict.')
        sys.exit(1)

try:
    # Python 2
    irange = xrange
except NameError:
    #Python 3
    irange = range


try:
    # Python 2
    fold = reduce
except NameError:
    import functools
    fold = functools.reduce

try:
    # Python 2.x and Python 3.2+
    is_callable = callable
except NameError:
    # Python 3.0 and Python 3.1
    from collections import Callable

    def is_callable(x):
        return isinstance(x, Callable)

try:
    # Python 3
    def _dummy():
        pass

    _dummy.__name__
    del _dummy

    def function_name(f):
        return f.__name__
except AttributeError:
    # Python 2
    def function_name(f):
        return f.func_name

try:
    # Python 2
    unicode()

    def has_unicode_type():
        return True

except NameError:
    # Python 3
    def has_unicode_type():
        return False

def itervalues(dictionary):
    '''Get an iterator over dictionary values.

    Attempts to avoid copying the dictionary values.

    Args:
        dictionary: The dictionary for which an iterator over values is wanted.

    Returns:
        An iterator over the dictionary values.
    '''
    try:
        return dictionary.itervalues()
    except AttributeError:
        pass
    return iter(dictionary.values())

def iteritems(dictionary):
    '''Get an iterator over dictionary items.

    Attempts to avoid copying the dictionary items.

    Args:
        dictionary: The dictionary for which an iterator over items is wanted.

    Returns:
        An iterator over the dictionary items.
    '''
    try:
        return dictionary.iteritems()
    except AttributeError:
        pass
    return iter(dictionary.items())

try:
    # Python 2.7+ and Python 3.2+
    from functools import total_ordering as totally_ordered
except ImportError:
    # Python 2.6 version from
    # http://code.activestate.com/recipes/576685-total-ordering-class-decorator/
    # This recipe doesn't actually work on Python 3.0 or 3.1 but that doesn't
    # matter since neither of those implementations cause the
    # extra methods added by the decorator to be exercised.  This may become
    # an issue in future if somebody creates an alternative Python 3
    # implementation which is not up to Python 3.2 completeness.  That seems
    # unlikely.
    def totally_ordered(cls):
        'Class decorator that fills-in missing ordering methods'
        convert = {
            '__lt__': [('__gt__', lambda self, other: other < self),
                       ('__le__', lambda self, other: not other < self),
                       ('__ge__', lambda self, other: not self < other)],
            '__le__': [('__ge__', lambda self, other: other <= self),
                       ('__lt__', lambda self, other: not other <= self),
                       ('__gt__', lambda self, other: not self <= other)],
            '__gt__': [('__lt__', lambda self, other: other > self),
                       ('__ge__', lambda self, other: not other > self),
                       ('__le__', lambda self, other: not self > other)],
            '__ge__': [('__le__', lambda self, other: other >= self),
                       ('__gt__', lambda self, other: not other >= self),
                       ('__lt__', lambda self, other: not self >= other)]
        }
        if hasattr(object, '__lt__'):
            roots = [op for op in convert if getattr(cls, op) is not getattr(object, op)]
        else:
            roots = set(dir(cls)) & set(convert)
        assert roots, 'must define at least one ordering operation: < > <= >='
        root = max(roots)       # prefer __lt __ to __le__ to __gt__ to __ge__
        for opname, opfunc in convert[root]:
            if opname not in roots:
                opfunc.__name__ = opname
                opfunc.__doc__ = getattr(int, opname).__doc__
                setattr(cls, opname, opfunc)
        return cls




'''Adding extension operators.'''

__author__ = 'Robert Smallshire'

from ._portability import function_name


def add_method(function, klass, name=None):
    '''Add an existing function to a class as a method.

    Note: Consider using the extend decorator as a more readable alternative
        to using this function directly.

    Args:
        function: The function to be added to the class klass.

        klass: The class to which the new method will be added.

        name: An optional name for the new method.  If omitted or None the
            original name of the function is used.

    Returns:
        The function argument unmodified.

    Raises:
        ValueError: If klass already has an attribute with the same name as the
            extension method.
    '''
    # Should we be using functools.update_wrapper in here?
    if name is None:
        name = function_name(function)
    if hasattr(klass, name):
        raise ValueError("Cannot replace existing attribute with method "
                         "'{name}'".format(name=name))
    setattr(klass, name, function)
    return function


def extend(klass, name=None):
    '''A function decorator for extending an existing class.

    Use as a decorator for functions to add to an existing class.

    Args:
        klass: The class to be decorated.

        name: The name the new method is to be given in the klass class.

    Returns:
        A decorator function which accepts a single function as its only
        argument.  The decorated function will be added to class klass.

    Raises:
        ValueError: If klass already has an attribute with the same name as the
            extension method.
    '''
    def decorator(f):
        return add_method(f, klass, name)

    return decorator



__author__ = 'Robert Smallshire'


class Record(object):
    '''A class to which any attribute can be added at construction.'''

    def __init__(self, **kwargs):
        '''Initialise a Record with an attribute for each keyword argument.

        The attributes of a Record are mutable and may be read from and written
        to using regular Python instance attribute syntax.

        Args:
            **kwargs: Each keyword argument will be used to initialise an
                attribute with the same name as the argument and the given
                value.
        '''
        self.__dict__.update(kwargs)

    def __eq__(self, rhs):
        return self.__dict__ == rhs.__dict__

    def __ne__(self, rhs):
        return self.__dict__ != rhs.__dict__

    def __str__(self):
        '''A string representation of the Record.'''
        return "Record(" + ', '.join(str(key) + '=' + str(value) for key, value in self.__dict__.items()) + ')'

    def __repr__(self):
        '''A valid Python expression string representation of the Record.'''
        return "Record(" + ', '.join(str(key) + '=' + repr(value) for key, value in self.__dict__.items()) + ')'


def new(**kwargs):
    '''A convenience factory for creating Records.

    Args:
        **kwargs: Each keyword argument will be used to initialise an
            attribute with the same name as the argument and the given
            value.

    Returns:
        A Record which has a named attribute for each of the keyword arguments.
    '''
    return Record(**kwargs)

